#!/usr/bin/perl
package POE::Component::goanseech::URLfetcher;
##############################################################################
=pod

=head1 NAME

POE::Component::goanseech::URLfetcher

=head1 DESCRIPTION

extracts the links from a given HTML file and
recusrively parses the files that are pointed inside 

=head1 AUTHOR

  Bruno Bttcher <bboett at adlp dot org>

=head1 COPYRIGHT and LICENCE

 Copyright (c) 2003 Bruno Boettcher

 POE::Component::goanseech::URLfetcher is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; version 2
 of the License.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


=cut

##############################################################################
use warnings;
use strict;
use HTML::TreeBuilder;
use Data::Dumper;

use POE;
use POE::Filter::Reference;
use POE::Component::Client::TCP;

use HTTP::Request::Common qw(GET POST);
use POE::Component::Client::HTTP;




##############################################################################
=pod

=item _stop

to keep this session alive set an alias

=cut
######################################################################
sub _stop
{
  #print("stopping extract torrents...\n");
  #exit(0);
}# sub _stop
##############################################################################
=pod

=item shutdown

close this session

=cut
######################################################################
sub shutdown
{
  my ($this, $kernel) = @_[OBJECT, KERNEL, ARG0 ];
  #print("shutdown extract torrents...\n");
  $_[KERNEL]->post("ua", "shutdown"); 
  $_[KERNEL]->alias_remove( $this->{"name"});
  exit();
}# sub shutdown
##############################################################################
=pod

=item CTOR

instantiation of the Helper

=cut

##############################################################################
sub new 
{
  my($class,$sessionname) = @_;
  my $this = bless {}, "POE::Component::goanseech::URLfetcher";
  $this->{"name"} = $sessionname;
  return $this;
}
##############################################################################
=pod

=item init

Initiaize, means set up this module

=cut

######################################################################
sub init
{
  my ($this,$settings,$arguments) = @_;

  $this->{"settings"} = $settings;

  $this->{"hrefs"} = {}; 
  $this->{"pagecount"} = 0;
  $this->{"indexRun"} = 1; # don't try to interpret the pages analyzed in this
                           # run, just extract the URL for further parsing

  $this->{"url_list"} = []; #start pages for this to search for
  $this->{"visitedURL"} = {};
  $this->{"problems"} = "";
  $this->{"nohosts"} = []; # hosts that make some sort of prblems and should be avoided

  foreach my $key (keys(%$arguments))
  {
    $this->{$key} = $arguments->{$key};
  }# foreach my $key (keys($arguments))


#read in the urls to monitor
  $settings->{"rootdir"} .= "." if(!($settings->{"rootdir"}));
  $settings->{"rootdir"} .= "/" if(!($settings->{"rootdir"} =~ /\/$/));

  if(open(FILE,$settings->{"rootdir"}."url.conf"))
  {
    foreach my $line (<FILE>)
    {
      chomp($line);
      #add the line only if it isn't a comment
      my $url_list = $this->{"url_list"};
      push(@$url_list,$line) if($line && !($line =~ /^#/));
    }# foreach my $line (<FILE>)
    close(FILE);
  }# if(open(FILE,"url.conf"))
  else
  {
    $this->print("couldn't open ".$settings->{"rootdir"}."url.conf\n");
  }# else
  #make a backup of this list ulterior use...
  my @tmp = @{$this->{"url_list"}};
  $this->{"index_list"} = \@tmp;

  $this->startSession();

}#sub init
##############################################################################
=pod

=item startSession

start our own session to be able to have timered stuff

=cut
######################################################################
sub startSession
{
  my ($this) = @_;
  my $settings = $this->{"settings"};

#start a http client
  POE::Component::Client::HTTP->spawn( 
      Alias	=> 'ua' ,
      Agent	=> 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0)',
      Timeout	=> $settings->{"webGetTmout"});


  POE::Session->create
    (
     object_states =>
     [      

     $this => { 
     _start         => "_start",
     _stop         => "_stop",
     "shutdown" => "shutdown",
     "got_webpage" => "got_webpage",
     "fetchURLs" => "fetchURLs",
     "reset" => "reset",
     "doneBatch" => "doneBatch",
     "clearNohosts"  => "clearNohosts",
     "checkWebTmout"  => "checkWebTmout",
     }
  ]
    );
}# sub startSession
##############################################################################
=pod

=item _start

to keep this session alive set an alias

=cut
######################################################################
sub _start
{
  my ($this, $kernel, $arg ) = @_[OBJECT, KERNEL, ARG0 ];
  $_[KERNEL]->alias_set($this->{"name"});
  $this->{"kernel"} = $kernel;
}# sub _start
##############################################################################
=pod

=item fetchURLs

work off the list and fetch the pages pointed by this list

=cut
######################################################################
sub fetchURLs
{
  my ($this, $kernel, $arg ) = @_[OBJECT, KERNEL, ARG0 ];
  #print("This is $this->{name} in fetchURLs \n");
  $this->{"pagecount"} = 0;
  #print("dunmmpoing array: ".Dumper($arg)."\nrype=".ref($arg)."\n");
  $this->{"url_list"} = $arg if($arg && ref($arg) eq 'ARRAY');
  my $sendcount = 0;

  my @clearedlist = ();
  foreach my $url (@{$this->{"url_list"}}) 
  {
    $this->print("posting $url\n");
    if(!($this->{"visitedURL"}->{"$url"}))
    {
      $this->{"visitedURL"}->{"$url"} = 1;
      $kernel->post( "ua" => "request", "got_webpage", GET $url);
      $sendcount++;
      push(@clearedlist,$url);
    }# if(!($this->{"visitedURL"}->{"$url"}))
    else
    {
      #my $num = scalar(@{$this->{"url_list"}});
      #@{$this->{"url_list"}} = grep((!/^\Q$url\E$/), @{$this->{"url_list"}});
      #my $newnum = scalar(@{$this->{"url_list"}});
      #print("fetchURLs: allready fetched $url reduced from $num to $newnum\n");
      $this->print("fetchURLs: allready fetched $url\n");
    }# else
  }

  $this->{"url_list"} = \@clearedlist;

  if($sendcount <= 0)
  {
    #we need to get back to the main controlling process if somethings went wrong here...
    $this->print("bailing out to $this->{urlEvent} on $this->{querySession}\n");
    $_[KERNEL]->post($this->{"querySession"}, $this->{"queryEvent"});
  }# if($sendcount <= 0)
  else
  {
    #print("Issued requests: ".Dumper($this->{"url_list"})."\n");
    my $tmout = $this->{"settings"}->{"webfetchtmout"}*60;
    if($tmout >0)
    {
      $kernel->delay("checkWebTmout" => $tmout);
    }# if($tmout >0)
  }# else
}# sub fetchURLs
######################################################################
=pod

=item got_webpage

got one of the webpages, parse them!

=cut
######################################################################
sub got_webpage 
{
  my ($this, $heap, $request_packet, $response_packet ) = @_[OBJECT, HEAP, ARG0, ARG1 ];
  $this->{"pagecount"}++;

  my $http_request  = $request_packet->[0];
  my $http_response = $response_packet->[0];

  #my $response_string = $http_response->as_string();
  #$response_string =~ s/^/| /mg;
  #print ",", '-' x 78, "\n";
  #print $response_string;
  #print "`", '-' x 78, "\n";
  #_content
  #print("RESPONCE: ".Dumper($http_response)."\n");
  $this->print("got page ".$this->{"pagecount"}."/".scalar(@{$this->{"url_list"}}));

  #starting parsing
  #print("page URI = ".ref($http_response-)."\n");
  #print("page URI = ".$http_response->base."\n");
  if($http_response->is_success)
  {
    my $base = $http_response->base;
    $base = $base->as_string();
    $this->get_hrefs($http_response->content_ref(), $base, $_[KERNEL]);
  }# if($http_response->is_success)
  else
  {
    my $msg = "there was a problem: "; 
    my $errortree = HTML::TreeBuilder->new;
    $errortree->parse($http_response->error_as_HTML());   # !
      $errortree->eof();
    $msg .= $errortree->as_text();
    $errortree->delete; # clear memory!
      $msg .= "  with  ".$http_response->{"_msg"} if($http_response->{"_msg"});
    $msg .= " for ".$http_request->uri() if($http_request->uri());
    $msg .= "\n";
    if($msg =~ /302 Found/i)
    {
      my $base = $http_response->base;
      $base = $base->as_string();
      $this->get_hrefs($http_response->content_ref(), $base, $_[KERNEL]);
    }# if($msg =~ /302 Found/i)
    elsif($this->{"settings"}->{"excludeBadURLs"} && ($msg =~ /Occurred 404 Not Found/i) ||($msg =~ /Occurred 500/i))
    {
      #add this hosts to the ones that should be avoided
      $this->print("ADDING ".$http_request->uri()." to nohosts for : $msg\n");
      push(@{$this->{"nohosts"}},$http_request->uri()) if($http_request->uri());
    }# if($msg =~ /Occurred 404 Not Found/i)
    #print($msg); 
    $this->{"problems"} .= $msg; 
  }# else

  if($this->{"pagecount"} >= scalar(@{$this->{"url_list"}}))
  {
    $this->print("page-batch complete\n");
    $this->switchToTorrentfetching($_[KERNEL]);
  }#if($this->{"pagecount"} >= scalar(@{$this->{"url_list"}}))

}# sub got_webpage 
######################################################################
=pod

=item switchToTorrentfecthing

we finished fetching torrents, now lets start the cycle to fetch the real files...


=cut
######################################################################
sub switchToTorrentfetching 
{
  my ($this,$kernel) = @_;
  $this->{"url_list"} = [];

  if($this->{"indexRun"})
  {
    $this->print("index run\n");
    #print("sendBatch keys: ".Dumper(\@keys));
    my $doubleRefs = {};
    my $hrefs = $this->{"hrefs"};
    foreach my $key (sort(keys(%$hrefs)))
    {
      #avoid adding the same link twice
      if(!($doubleRefs->{$hrefs->{$key}}))
      {
	$doubleRefs->{$hrefs->{$key}} = 1;	
	my $url_list = $this->{"url_list"};
	push(@$url_list,$hrefs->{$key});
      }# if(!($doubleRefs->{$hrefs->{$key}}))
    }# for

    #next run is for real
    $this->{"indexRun"} = 0;

    if(scalar(@{$this->{"url_list"}}) > 0)
    {
      $this->print("index run posting new fetchURL's\n");
      $kernel->yield("fetchURLs");
      return;
    }
  }# if($this->{"indexRun"})

  $this->print($this->{"problems"}) if($this->{"problems"}); 
  $this->print("posting $this->{urlEvent} on $this->{querySession}\n");
  #prepare jobs for series manager...
  #
  my @request = ( );
  #reset the listing of pages to seach, since we are asking for a whole new bunch of pages...
  my @keys = sort(keys(%{$this->{"hrefs"}}));
  #print("sendBatch keys: ".Dumper(\@keys));
  foreach my $key (@keys)
  {
    #print("added request exile to list: $key\n") if($key =~ /last.*?exile/i);
    push(@request, [ "find", 
	{ "line" =>$key, "url" => $this->{"hrefs"}->{$key}}
	]);
  }# for
  my $job = {
    "requests" => \@request,
    "session" => $this->{"name"},
    "event"	=> "doneBatch",
    "tmout"	=> time()+5*60,
  };
  $this->print("got_webpage exiting now posting ".$this->{"queryEvent"}." to ".$this->{"querySession"}."\n");
  $kernel->post($this->{"querySession"}, $this->{"queryEvent"}, $job);
}#sub switchToTorrentfecthing 
##############################################################################
=pod

=item get_hrefs

analyses a given html file and extracts all the HREFS, sends back a hash with
the comment as key and the URL as value

=cut

######################################################################
sub get_hrefs 
{
  my ($this,$content, $baseuri,$kernel) = @_;
  my $tree = HTML::TreeBuilder->new;
  $tree->parse(${$content});
  $tree->eof();
  if(ref($baseuri) && ref($baseuri) ne 'SCALAR' )
  {
    die("HELLLL whats this: ".Dumper($baseuri)."\n");
  }# if(ref($baseuri) && ref($baseuri) ne 'SCALAR' )
  if(!($baseuri =~ /\/$/))
  {
    if($baseuri =~ /(^.*?\/).*?\.s?html?$|(^.*?\/).*?\.php$|(^.*?\/).*?\.php\?.*$|(^.*?\/).*?\.torrent$/)
    {
      if($baseuri =~ /\.torrent$/)
      {
	      $this->{"hrefs"}->{$baseuri} = $baseuri;
      }
    }# if($baseuri =~ /\.html?$|\.php$/)
    else
    {
      $baseuri = $baseuri."/"; 
      $this->print("corrected basuri to $baseuri\n");
    }
  }

  #print( "Document = ".${$_[0]}."\n" );


  #my @tags = $tree->extract_links(['a']);  # !
  my @tags = $tree->look_down('_tag', 'A');  # !
    #my @tags = $tree->find_by_tag_name('A');  # !
    #if(@tags) {
      #  $heading = $h1->as_text;   # !
      #} else {
	#  warn "No heading in $_[0]?";
	#}
    # print("the results : ".Dumper($tree)."\n");
    my %result = ();
  foreach my $line (@tags)
  {
    my  $heading = $line->as_text;   # !
      #print("one line : ".Dumper($heading->content_list)."\n");
      #print("one line : ".Dumper($line)."\n");
      #print("one line : ".Dumper($line->all_attr_names())."\n");
      #    one line : $VAR1 = '_parent';
      #  $VAR2 = '_content';
      #  $VAR3 = 'href';
      #  $VAR4 = '_tag';

      #global buildup!
      my $retrievedURL = $line->attr('href');
    if($retrievedURL)
    {
      if($line->attr('_content'))
      {

	my $retrievedComment = shift(@{$line->attr('_content')});
	print("found redirect with $retrievedComment for $retrievedURL\n") if($retrievedComment eq "here");

	my $altURL;
	if(!($retrievedURL =~ /^http/))
	{
	  #print("LINE ".$retrievedURL." with no valid hostpart.... :".Dumper(($line->all_attr_names()))."\nbasuri = ".$baseuri."\n");
	  #we want to avoid popping out the host part, so remove the known / at the beginning of the url
	  $baseuri =~ /http:\/\/(.*)$/;
	  my $suburi = $1;
	  #building up a step higher
	  my @path = split('/',$suburi);
	  pop(@path);
	  #if we popped out the host instead of a path component, forget the proceding
	  if(@path)
	  {
	    #don't accept URL's with ? in it...
	    if(!($retrievedURL=~ /\?/))
	    {
	      my $altbase;
	      $altbase = "http://".join('/',@path);
	      if($altbase =~ /\/$/)
	      {
		$altURL = $altbase.$retrievedURL;
	      }# if($altbase =~ /\/$/)
	      else
	      {
		$altURL = $altbase."/".$retrievedURL;
	      }
	      #ARGH mircx adaptation....
	      $altURL =~ s/\/\/torrent/\/torrent/g;
	      $altURL =~ s/\/torrent\/torrent/\/torrent/g;
	    }# if(!($retrievedURL=~ /\?/))
	  }# if(@path)
	  #print("adding alternative URL for ".$retrievedComment." at $baseuri\n");
	  #print("  for $altURL for original $retrievedURL\n");
	  #we have an URL with an absolute patrh name, thus we need the url without any other path, the raw hostname...
	  if($retrievedURL =~ /^\//)
	  {
	    $baseuri =~ /^(http:\/\/.*?)\//;
	    $retrievedURL = $1.$retrievedURL ;
	  }# if($baseuri =~ /\/$/ && $retrievedURL =~ /^\//)
	  else
	  {
	    $retrievedURL = $baseuri.$retrievedURL;
	  }
	  #print("  giv $retrievedURL\n  and $altURL\n");
	}# if(!($line->attr('href') =~ /^http/))

	#http://anime.mircx.com/ adaptation, they made a really stupid setup for their webpage....
	#
        if($retrievedURL =~ /^(http:\/\/anime\.mircx\.com\/)torrent\/(.*)/)
	{
	  $retrievedURL = $1."download/".$2; 
	}#if($retrievedURL =~ /^(http:\/\/anime\.mircx\.com\/)torrent\/(.*)/)

        if($altURL && $altURL =~ /^(http:\/\/anime\.mircx\.com\/)torrent\/(.*)/)
	{
	  $altURL = $1."download/".$2; 
	}#if($retrievedURL =~ /^(http:\/\/anime\.mircx\.com\/)torrent\/(.*)/)

	#some pages truncate the name of the comment, but the final name in the torrentlink is correct, thus swap them
	$retrievedComment = $retrievedURL if($retrievedURL =~ /torrent\s*$/i);
	if(grep((/^\Q$retrievedURL\E$/), @{$this->{"nohosts"}}))
	{
	  #$this->print("Problem URL, ignoring $retrievedURL\n");
	  $this->print("problem url exile: $retrievedURL\n") if($retrievedURL =~ /last.*?exile/i);
	}# if(grep((/\Q$retrievedURL\E/), @{$this->{"nohosts"}}))
	elsif($retrievedURL =~ /mailto:/i)
	{
	  #$this->print("crap URL, ignoring $retrievedURL\n");
	}# if(grep((/\Q$retrievedURL\E/), @{$this->{"nohosts"}}))
	else
	{
	  $this->{"hrefs"}->{$retrievedComment} = $retrievedURL;
	  if($altURL)
	  {
	    my $newcomment = $retrievedComment."B";
	    $this->{"hrefs"}->{$newcomment} = $altURL ;
	  }
	  $this->print("added url exile to list: $retrievedURL\n") if($retrievedURL =~ /last.*?exile/i);
	}# else
      }# if($line->attr('_content'))
      else
      {
	my $msg = "problems while fetching $baseuri no content ";
	if ($this->{"settings"}->{"excludeBadURLs"} && $retrievedURL)
	{
	  $msg .= "for $retrievedURL"; 
	  #$this->print("ADDING $retrievedURL to nohosts for : $msg\n");
	  push(@{$this->{"nohosts"}},$retrievedURL);
	}
	$msg .= "\n";
	#$this->print($msg);
	$this->{"problems"} .= $msg; 
      }# else
    }#if($retrievedURL)
    #$this->print("adding ".$line->attr('href')." with key ".$line->attr('_content')."\n");
    #$this->print("one line : ".shift(@{$line->attr('_content')}).":".$line->attr('href')."\n");
    #$this->print("one line : ".ref(\$heading)."\n");
  }# foreach my $line (@$hrefs)
  $tree->delete; # clear memory!
    #return \%result;
}# sub get_hrefs 
##############################################################################
=pod

=item reset

put this object back into boot state, mainly necessary for purging the allready visited pages DB

=cut
######################################################################
sub reset
{
  my ($this, $kernel, $arg ) = @_[OBJECT, KERNEL, ARG0 ];
  $this->{"hrefs"} = {}; 
  $this->{"pagecount"} = 0;
  $this->{"indexRun"} = 1; 
  $this->{"url_list"} = [];
  $this->{"visitedURL"} = {};
  $this->{"problems"} = "";
}# sub reset
##############################################################################
=pod

=item doneBatch

The series server finished parsing a batch of lines, if 

=cut
######################################################################
sub doneBatch
{
  my ($this, $kernel, $arg ) = @_[OBJECT, KERNEL, ARG0 ];

  #push(@{$this->{"actjob"}->{"result"}}, $hash);
  my $torrents = [];
  my $url_list = [];
  foreach my $hash (@$arg)
  {
    $this->print("doneBatch processing ".Dumper($hash)."\n");
    if($hash->{"url"} =~ /^(.*)\.torrent\s*$/i)
    {
      #we have a ready torrent file that was selected!
      #doubles are avoided by using the series pattern as hash key
      #now we have ot check if it is of the right type!

      my $reduced = $1;
      $this->print("doneBatch: reduced = $reduced\n");
      if($reduced =~ /$this->{"settings"}->{"typestoavoid"}$/i)
      {
	$this->print("WRONGTYPE for ".$hash->{"url"}."\n");
      }# if($reduced =~ /$this->{"settings"}->{"typestofetch"}$/i)
      else
      {
        push(@{$torrents}, $hash);
        $this->print("TORRENT ".$hash->{"pattern"}." @ ".$hash->{"url"}."\n");
      }# else
    }# if($hash->{"url"} =~ /torrent\s*$/i)
    elsif($hash->{"url"} =~ /^\s*http/i)
    {
      #TODO don't know why its necessary but....
     while($hash->{"url"} =~ /(http:\/\/.*?)\/\/(.*)/)
     {
         $hash->{"url"} = $1."/".$2;
     }# while($hash->{"url"} =~ /(http:\/\/.*?)\/\/(.*)/)
      $this->print("LINK ".$hash->{"pattern"}." @ ".$hash->{"url"}."\n");
      if($this->{"visitedURL"}->{$hash->{"url"}})
      {
	$this->print("allready visited...\n");
      }# if($this->{"visitedURL"}->{"$url"})
      elsif($hash->{"url"} =~ /file\?info_hash/i)
      {
	#$this->print("junk...\n");
      }# if($this->{"visitedURL"}->{"$url"})
      else
      {
	my $url = $hash->{"url"};
	my $count = ($url =~ tr/?//);
	$this->print("CHAR COUNT $count\n");

	if($count >1)
	{
	  $this->print("where does this junk come from=??? ".$url."\n");
	}# if($count >1)
	#add the url only if we haven't it allready in the list
	elsif(!(grep(/\Q$url\E/,$url_list)) && !($url =~ /mailto:/i))
	{
	  push(@{$url_list},$hash->{"url"}); 
	}
      }
    }# else
    else
    {
      $this->print("UNKNOWN ".$hash->{"pattern"}." @ ".$hash->{"url"}."\n");
    }# else
  }#// foreach my $hash ($arg)

  if(scalar(@{$url_list}) > 0)
  {
    $this->print("we still have some URL's to parse:".scalar(@{$url_list})."\n");
    $this->print("posting $this->{urlEvent} to $this->{urlSession}");
    $kernel->yield("fetchURLs", $url_list); 
  }# if(scalar($url_list) > 0)
  else
  {
    $this->print("no more URL's to visit... num of torrents: ".scalar(@{$torrents})."\n");
    $kernel->post($this->{"torrSession"}, $this->{"torrEvent"},$torrents); 
  }# else
}# sub doneBatch
######################################################################
=pod

=item clearNohosts

riding on an adls connection sometimes lets cast of all the URL_s whilst a
disconnect marking them all as invalid.... thus at reapeated intervals we clear
the blacklist of hosts

invalidate if the tmout is set to 0, set in settings->dropErrorHostsIn mesure
are hours

=cut

######################################################################
sub clearNohosts 
{
  my ($this, $kernel, $heap,$arg) = @_[OBJECT, KERNEL, HEAP,ARG0];
  $this->{"nohosts"} = []; 
  my $tmout = $this->{"settings"}->{"dropErrorHostsIn"}*60*60;
  if($tmout >0)
  {
    $kernel->delay("clearNohosts" => $tmout);
  }# if($tmout >0)
}# sub clearNohosts 
######################################################################
=pod

=item checkWebTmout

strangely enough sometimes a calls goes lost.... don't know why....
enfore the torrent parsing after a given timout

=cut
######################################################################
sub checkWebTmout 
{
  my ($this, $heap, $request_packet, $response_packet ) = @_[OBJECT, HEAP, ARG0, ARG1 ];
  if(scalar(@{$this->{"url_list"}}) > 0)
  {
    $this->print("Error!! the call to certain webpages didn't return even after a timout\n");
    $this->switchToTorrentfetching($_[KERNEL]);
  }#if($this->{"pagecount"} >= scalar(@{$this->{"url_list"}}))
}# sub checkWebTmout 
#########################################################
=pod

=item print

delegator for printing

=cut

#########################################################
sub print
{
  my ($this,$msg) = @_;

  if($this->{"kernel"})
  {
    $this->{"kernel"}->post($this->{"querySession"}, "print", $msg,"urlfetcher") 
  }
  else
  {
    print($msg);
  }# else
}# sub print
1
__END__

=back


=head1 SEE ALSO

http://poe.perl.org/

=cut

