##############################################################################
=pod

=head1 NAME

series Parser

=head1 DESCRIPTION

XML parser to read in the series definition for the series server

=head1 AUTHOR

Bruno BTTCHER <bboett at adlp.org>

=head1 COPYRIGHT AND LICENCE

Copyright (c) 2003 Bruno Boettcher


  seriesparser.pm is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; version 2
  of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=head1 Methods of this class

=over

=cut

##############################################################################
package POE::Helper::Leve::seriesparser;
use strict;

use Data::Dumper;
use XML::Parser;
#use ObjectTemplate;
#@ISA = ("ObjectTemplate");
#attributes("sysdata");

my $seriesTree = {};
my @refStack = ();
my $tokenName ="";
my $actReply ="";
my $actText ="";

##############################################################################
=pod

=item CTOR

instantiation of the seriesparser

=cut

##############################################################################
sub new {
  my $this = bless {}, "POE::Helper::Leve::seriesparser";
  return $this;
}
##############################################################################
=pod

=item init

initialisation of the seriesparser

=cut

##############################################################################
sub init
{
  my ($this) = @_;
  #print "seriesparser::".$this->isa()." init\n";
  $this->{"parser"} = new XML::Parser(
      #Style => 'Debug',
      #Style => 'Subs',
      Style => 'Subs',
      Pkg => 'POE::Helper::Leve::seriesparser',
      ErrorContext => 3,  #num of lines around the error
      );
  $this->{"parser"}->setHandlers(Char    => \&text,
                                Default => \&other);

  return 0;
}#sub init
##############################################################################
=pod

=item parse

takes a filename and the 2 references to the translation tables,  and parses
it, you need 2 since you surely want to go in both directions, means tag to
message, and trigger to tag... you can use both independently, since they share
the actual languge definition hash...  don't forget to initalize the refs since
the sub fills them, if there were undef, they will be lost...

=cut

##############################################################################
sub parse
{
  my ($this,$filename,$tTree) = @_;
  
  $seriesTree = {};

  $seriesTree = $tTree if($tTree);

  my $parser = $this->{"parser"};
  $this->{"patternNames"} = join(' ',keys(%{$seriesTree}));
  #print("seriesParese: start parse $filename\n");
  $parser->parsefile($filename);
  #print("seriesParese: tokens ".Dumper(keys(%{$seriesTree}))."\n");
  return $seriesTree;
}#sub shutdown
##############################################################################
=pod

=item isa

issue the type of this object

=cut

##############################################################################
sub isa
{
  return "seriesparser";
}#sub isa
##############################################################################
=pod

=item version

issue the version of this module

=cut

##############################################################################
sub version
{
  return '$Revision: 1.1 $';
}#sub isa
#########################################################
=pod

=item debug

issue debugging messages to the console

=cut

#########################################################
sub debug
{
  my ($this,$msg) = @_;
  $msg = $this if(!$msg);
  print("$msg\n");
}# sub debug
#########################################################
=pod

=item series

Callback for an opening series tag

Is the first useful tag of the catalog env, since it holds the definition for
one type of series

=cut

#########################################################
sub series
{
  #print "start of series : ".Dumper(@_)."\n"; 
  my ($expat, $envname, $tok,$tag) = @_;
  $actTok = {};
  shift;shift;
  while(@_)
  {
    my $tok = shift,
    my $arg = shift;
    $actTok->{$tok} = $arg;
  }# while(@_)
  $tag = $actTok->{"name"};
  $tag = $this->patternrevision($tag);
  $tokenName = $tag;
  print "start of series : $tag\n"; 
  #$seriesTree->{$tag} = {} if(!($seriesTree->{$tag}));
  #$actTok = $seriesTree->{$tag};
  $actTok->{"priority"} = 0 if(!($actTok->{"priority"}));
  if($actTok->{"fetched"})
  {
    my %subfetched = ();
    foreach my $i (split(' ',$actTok->{"fetched"}))
    {
      $i = int($i); #force numerical interpretation
	$subfetched{$i} = 1;
    }#foreach my $i (split(' ',$loaded))
    $actTok->{"intermediate"} = \%subfetched;
    delete $actTok->{"fetched"};
  }# if($actTok->{"fetched"})

  if($seriesTree->{$tag})
  {
    #the given series is allready in our base, check for differences
    if($seriesTree->{$tag}->{"min"} < $actTok->{"min"})
    {
      $seriesTree->{$tag}->{"min"} = $actTok->{"min"}; 
    }
    $seriesTree->{$tag}->{"max"} = $actTok->{"max"} if($seriesTree->{$tag}->{"max"} < $actTok->{"max"});
    $seriesTree->{$tag}->{"priority"} = $actTok->{"priority"} if($seriesTree->{$tag}->{"priority"} != $actTok->{"priority"});

    foreach my $subEpisode (keys(%{$actTok->{"intermediate"}}))
    {
      if(!($seriesTree->{$tag}->{"intermediate"}->{$subEpisode}))
      {
	$seriesTree->{$tag}->{"intermediate"}->{$subEpisode} = 1;
      }# if(!($series{$pattern}->{"intermediate"}->{$subEpisode}))
    }# foreach my $subEpisode (keys(%{$actTok->{"intermediate"}}))
  }# else
  else
  {
    $seriesTree->{$tag} = $actTok;
  }# else
  #remove the actual pattern from our list
  $this->{"patternNames"} =~ s/\Q$tag\E //;
}# sub series
#########################################################
=pod

=item alias

Callback for an opening alias tag

This one defines an alias for an existing record, all changes mad through this one are mirrored back to the original record.

=cut

#########################################################
sub alias
{
 #print "start of series : ".Dumper(@_)."\n"; 
  my ($expat, $envname, $tok,$tag) = @_;
  $tag = $this->patternrevision($tag);
  $tokenName = $tag;
  #print "start of series : $tag\n"; 
  $seriesTree->{$tag} = {} if(!($seriesTree->{$tag}));
  $actTok = $seriesTree->{$tag};
  shift;shift;shift;shift;
  while(@_)
  {
    my $tok = shift,
    my $arg = shift;
    $actTok->{$tok} = $arg;
  }# while(@_)
  $actTok->{"type"} = "alias";
}# sub alias
#########################################################
=pod

=item catalog


this is invoqued when the starting tag of the catalog env is called

=cut

#########################################################
sub catalog
{
  #print("begin parsing:".Dumper($seriesTree)."\nend\n");
}# sub catalog
#########################################################
=pod

=item catalog_


this is invoqued when the ending tag of the catalog env is called

=cut

#########################################################
sub catalog_
{
  #print("finished parsing:".Dumper($seriesTree)."\nend\n");
  #print("finished parsing dumping langtree:".Dumper($langTree)."\nend\n");

  my $patternNames = $this->{"patternNames"};
    $patternNames =~ s/\s+/ /g;
    my @obsoletePatterns = split(/ /,$patternNames);
    foreach my $toRemove (@obsoletePatterns)
    {
      delete $this->{"series"}->{$toRemove};
    }# foreach my $toRemove (@obsoletePatterns)
}# sub catalog_
#########################################################
=pod

=item text

some raw text comes in, store it and let the elements handle that...

=cut

#########################################################
sub text
{
  my ($expat, $data) = @_;
  $data =~ s/
/ /g;
  $data =~ s/\s+/ /g;
  #if(!($data =~ /^\s*$/))
  #{
  #  print("PARSER::text with '$data' '$expat'\n");
  #}
  $actText .= "$data ";
}# sub text
#########################################################
=pod

=item other

some unkown text comes in, store it and let the elements handle that...

=cut

#########################################################
sub other
{
  my ($expat, $data) = @_;
  #print("PARSER::other with '$data' '$expat'\n");
  $data =~ s/
//g;
  $actText .= "O$data O " if(!($data =~ /^\s*$/));
}# sub other

1
######################################################################
=pod

=item patternrevision

switch pattern active chars to perlish and non greedy patterns

=cut

######################################################################
sub patternrevision
{
  my ($this,$line) = @_;
  #change * into .*?
  $line =~ s/(\w)\*/$1.*?/g;
  #change .* into .*?
  $line =~ s/\.\*(?!\?)/.*?/g;
  return $line;
}#sub patternrevision
__END__


=back

=head1 SEE ALSO

zebot home page  http://www.freesoftware.fsf.org/zebot/ 
Net::IRC, 
RFC 1459,
http://www.irchelp.org/, 
http://poe.perl.org/
http://www.cs.cmu.edu/~lenzo/perl/, 
http://www.infobot.org/,


=cut

